/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include "threads.h"

#ifdef MBTK_HAVE_THREAD_PTHREAD

#include <sys/stat.h>
#include <errno.h>

int mbtk_be_thread_create(mbtk_be_thread_t *thr, void *(*func)(void *), void *udata)
{
	if (pthread_create(&thr->thread, NULL, func, udata) != 0)
		return -1;
	return 0;
}

int mbtk_be_thread_wait(mbtk_be_thread_t *thr)
{
	void *res;
	if (pthread_join(thr->thread, &res) != 0)
		return -1;
	return 0;
}

mbtk_be_sem_t mbtk_be_semget(int init_cnt)
{
	int sem = semget(IPC_PRIVATE, 1, IPC_CREAT | S_IRUSR | S_IWUSR);
	if (sem < 0)
		return MBTK_BE_INVALID_SEM;

	if (semctl(sem, 0, SETVAL, init_cnt) < 0) {
		semctl(sem, 0, IPC_RMID);
		return MBTK_BE_INVALID_SEM;
	}

	return sem;
}

void mbtk_be_semdel(mbtk_be_sem_t sem)
{
	semctl(sem, 0, IPC_RMID);
}

int mbtk_be_semwait(mbtk_be_sem_t sem, int block)
{
	struct sembuf sop;

	sop.sem_num = 0;
	sop.sem_op = -1;
	sop.sem_flg = block ? 0 : IPC_NOWAIT;

	for(;;) {
		if (semop(sem, &sop, 1) == 0)
			return 0;
		if (errno != EINTR) /* continue after EINTR, return error for anything else */
			return -1;
	}

	return -1;
}


int mbtk_be_sempost(mbtk_be_sem_t sem)
{
	struct sembuf sop;

	sop.sem_num = 0;
	sop.sem_op = 1;
	sop.sem_flg = 0;

	if (semop(sem, &sop, 1) == 0)
		return 0;

	return -1;
}

#endif


#ifdef MBTK_HAVE_THREAD_WIN32

/* need to wrap because of the different return value */
static DWORD mbtk_be_thread_wrap(void *thr_)
{
	mbtk_be_thread_t *thr = thr_;
	thr->func(thr->udata);
	return 0;
}

int mbtk_be_thread_create(mbtk_be_thread_t *thr, void (*func)(void *), void *udata)
{
	thr->func = func;
	thr->udata = udata;
	CreateThread(NULL, 0, wtr_os_thread_entry_wrap, thr, 0, NULL))
	return 0;
}

mbtk_be_sem_t mbtk_be_semget(int init_cnt)
{
	return CreateSemaphore(NULL, init_cnt, 0x7fffffe, NULL);
}

void mbtk_be_semdel(mbtk_be_sem_t sem)
{
	CloseHandle(sem);
}

int mbtk_be_semwait(mbtk_be_sem_t sem, int block)
{
	if (WaitForSingleObject(sem, (block ? INFINITE : 0)) == WAIT_OBJECT_0)
		return 0;

	return -1;
}

int mbtk_be_sempost(mbtk_be_sem_t sem)
{
	if (ReleaseSemaphore(sem, 1, 0) == 0)
		return 0;

	return -1;
}

#endif

