/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <libmbtk/src_3rd/genht/hash.h>
#include <assert.h>
#include "font.h"

void mbtk_sdl2_font_init(mbtk_sdl2_ttfcache_t *font_cache)
{
	static int lib_inited = 0;
	if (!lib_inited) {
		TTF_Init();
		lib_inited = 1;
	}
	font_cache->cache = htsp_alloc(strhash, strkeyeq);
}

void mbtk_sdl2_font_uninit(mbtk_sdl2_ttfcache_t *font_cache)
{
	htsp_entry_t *e;

	for (e = htsp_first(font_cache->cache); e; e = htsp_next(font_cache->cache, e)) {
		mbtk_sdl2_font_t *f = e->value;
		TTF_CloseFont(f->ttf);
		free(f->id);
		free(f);
	}

	htsp_free(font_cache->cache);
}


mbtk_sdl2_font_t *mbtk_sdl2_font_open(mbtk_sdl2_ttfcache_t *font_cache, const char *font_file, int size, int hinting, int style)
{
	char *id = malloc(strlen(font_file) + 32);
	mbtk_sdl2_font_t *font;

	sprintf(id, "%x-%d-%d-%s", size, hinting % 9, style % 9, font_file);
	font = htsp_get(font_cache->cache, id);
	if (font == NULL) {
		TTF_Font *ttf;
		ttf = TTF_OpenFont(font_file, size);
		if (ttf == NULL)
			return NULL;
		font = malloc(sizeof(mbtk_sdl2_font_t));
		font->ttf = ttf;
		font->id = id;
		font->size = size;
		TTF_SetFontHinting(font->ttf, hinting);
		TTF_SetFontStyle(font->ttf, style);
		htsp_set(font_cache->cache, id, font);
	}
	else
		free(id);

	return font;
}


mbtk_sdl2_font_ctx_t *mbtk_sdl2_font_ctx_init(SDL_Renderer *renderer, mbtk_sdl2_ttfcache_t *font_cache)
{
	mbtk_sdl2_font_ctx_t *ctx;
	ctx = malloc(sizeof(mbtk_sdl2_font_ctx_t));
	ctx->renderer = renderer;
	ctx->font_cache = font_cache;
	ctx->texture_cache = htsp_alloc(strhash, strkeyeq);
	ctx->ttf_solid = 0;
	return ctx;
}

void mbtk_sdl2_font_ctx_uninit(mbtk_sdl2_font_ctx_t *ctx)
{
	htsp_entry_t *e;

	for (e = htsp_first(ctx->texture_cache); e; e = htsp_next(ctx->texture_cache, e)) {
		mbtk_sdl2_text_t *txt = e->value;
		txt->refco = 1;
		txt->texture = NULL;
		mbtk_warning("mbtk_sdl2_font_ctx_uninit(): text with refco>0\n");
		mbtk_sdl2_text_destroy(ctx, txt);
	}
	htsp_free(ctx->texture_cache);
	free(ctx);
}

static SDL_Texture *mbtk_render_text_(SDL_Renderer *renderer, const char *message, mbtk_sdl2_font_t *font, SDL_Color color, int solid)
{
	SDL_Surface *surf;
	SDL_Texture *texture;

	/* empty string should result in an empty texture the cheap way */
	if (*message == '\0')
		return SDL_CreateTexture(renderer, SDL_PIXELFORMAT_RGB888, SDL_TEXTUREACCESS_STATIC, 0, 0);

	/* We need to first render to a surface as that's what TTF_RenderText returns, then load that surface into a texture */
	if (solid)
		surf = TTF_RenderText_Solid(font->ttf, message, color);
	else
		surf = TTF_RenderText_Blended(font->ttf, message, color);


	if (surf == NULL) {
		TTF_CloseFont(font->ttf);
		font->ttf = NULL;
		return NULL;
	}

	texture = SDL_CreateTextureFromSurface(renderer, surf);
/*	if (texture == NULL)
		fprintf(stderr, "font err 3\n");*/

	/* Clean up the surface and font */
	SDL_FreeSurface(surf);

	return texture;
}


mbtk_sdl2_text_t *mbtk_sdl2_text_new(mbtk_sdl2_font_ctx_t *ctx, mbtk_sdl2_font_t *font, const char *message, SDL_Color color)
{
	char *id, *msg_start;
	mbtk_sdl2_text_t *txt;
	int tlen = strlen(message);
	Uint32 clr;

	id = malloc(tlen + 40); /* if everything is 64 bits, we need 2*16+sep characters for the prefix */
	if (id == NULL)
		return NULL;
	memcpy(&clr, &color, sizeof(clr));
	msg_start = id + sprintf(id, "%p-%x-", (void *)font, clr);
	memcpy(msg_start, message, tlen+1);
	txt = htsp_get(ctx->texture_cache, id);

	if (txt == NULL) {
		txt = malloc(sizeof(mbtk_sdl2_text_t));
		txt->id = id;
		txt->font = font;
		txt->color = color;
		txt->message = msg_start;
		txt->texture = mbtk_render_text_(ctx->renderer, message, font, color, ctx->ttf_solid);
		txt->refco = 1;
		htsp_set(ctx->texture_cache, txt->id, txt);
	}
	else {
		txt->refco++;
		free(id);
	}

	return txt;
}

void mbtk_sdl2_text_destroy(mbtk_sdl2_font_ctx_t *ctx, mbtk_sdl2_text_t *txt)
{
	txt->refco--;
	assert(txt->refco >= 0);
	if (txt->refco == 0) {
		htsp_pop(ctx->texture_cache, txt->id);
		free(txt->id);
		SDL_DestroyTexture(txt->texture);
		free(txt);
	}
}


void mbtk_sdl2_text_render(mbtk_sdl2_font_ctx_t *ctx, mbtk_sdl2_text_t *txt, int x, int y)
{
	SDL_Rect dst;
	SDL_QueryTexture(txt->texture, NULL, NULL, &dst.w, &dst.h);
		
	dst.x = x;
	dst.y = y;

	SDL_RenderCopy(ctx->renderer, txt->texture, NULL, &dst);
}


#define SKIPEP    while((*curr == '=') || (*curr == ',') || (isspace(*curr))) curr++

static mbtk_sdl2_font_t *sdl2_resolve_font(sdl2_display_t *sd, const char *cfg)
{
	char *name, *curr;
	int style = 0, size = 16, hinting = 2, len;
	mbtk_sdl2_font_t *f;

	curr = strchr(cfg, ',');
	if (curr == NULL)
		len = strlen(cfg);
	else
		len = curr - cfg;

	name = malloc(len+1);
	memcpy(name, cfg, len);
	name[len] = '\0';

	while(curr != NULL) {
		SKIPEP;
		if (*curr == '\0') break;
		if (strncmp(curr, "size", 4) == 0) {
			curr += 4;
			SKIPEP;
			size = strtol(curr, &curr, 10);
		}
		else if (strncmp(curr, "bold", 4) == 0) {
			curr += 4;
			SKIPEP;
			style |= TTF_STYLE_BOLD;
		}
		else if (strncmp(curr, "italic", 6) == 0) {
			curr += 6;
			SKIPEP;
			style |= TTF_STYLE_ITALIC;
		}
		else if (strncmp(curr, "hinting", 7) == 0) {
			curr += 7;
			SKIPEP;
			hinting = strtol(curr, &curr, 10);
		}
	}
	f = mbtk_sdl2_font_open(&sd->font_cache, name, size, hinting, style);
	free(name);
	return f;
}

mbtk_font_id_t mbtk_sdl2_resolve_font(mbtk_display_t *display, int argc, char *argv[])
{
	int n;
	mbtk_font_id_t res;
	sdl2_display_t *sd = (sdl2_display_t *)display;

	for(n = 0; n < argc; n++) {
		mbtk_sdl2_font_t *f = sdl2_resolve_font(sd, argv[n]);
		if (f != NULL) {
			res.p = f;
			return res;
		}
	}

	res.l = -1;
	return res;
}

