/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <X11/Xutil.h>

#define MAYBE_SKIP_RENDER(skip_stmt) \
do { \
	xlib_win_t *sxw = (xlib_win_t *)win; \
	if (sxw->skip_render) { skip_stmt; } \
} while(0)

#define OX (-win->content_x0)
#define OY (-win->content_y0)

static int mbtk_xlib_draw_color(mbtk_window_t *win, int r, int g, int b, int a)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	XColor clr;
	unsigned long px;
	int succ;

	MAYBE_SKIP_RENDER(return 0);

	if (mbtk_xlib_clr_cache_get(xd, &px, r, g, b)) {
		XSetForeground(xd->display, xw->gc, px);
		return 0;
	}

	clr.red = (unsigned int)r << 8;
	clr.green = (unsigned int)g << 8;
	clr.blue = (unsigned int)b << 8;
	clr.flags = DoRed | DoGreen | DoBlue;
	succ = XAllocColor(xd->display, xd->cmap, &clr);

	if (!succ)
		return -1;

	mbtk_xlib_clr_cache_set(xd, clr.pixel, r, g, b);

	XSetForeground(xd->display, xw->gc, clr.pixel);
	return 0;
}

static void mbtk_xlib_draw_fillrect(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	XFillRectangle(xd->display, xw->window, xw->gc, x1 + OX, y1 + OY, x2 - x1 + 1, y2 - y1 + 1);
}

static void mbtk_xlib_draw_line(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	XDrawLine(xd->display, xw->window, xw->gc, x1 + OX, y1 + OY, x2 + OX, y2 + OY);
}

static void mbtk_xlib_draw_hline(mbtk_window_t *win, int x1, int y1, int x2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (x1 < x2)
		XDrawLine(xd->display, xw->window, xw->gc, x1 + OX, y1 + OY, x2 + OX, y1 + OY);
	else
		XDrawLine(xd->display, xw->window, xw->gc, x2 + OX, y1 + OY, x1 + OX, y1 + OY);
}

static void mbtk_xlib_draw_vline(mbtk_window_t *win, int x1, int y1, int y2)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (y1 < y2)
		XDrawLine(xd->display, xw->window, xw->gc, x1 + OX, y1 + OY, x1 + OX, y2 + OY);
	else
		XDrawLine(xd->display, xw->window, xw->gc, x1 + OX, y2 + OY, x1 + OX, y1 + OY);
}

int mbtk_xlib_draw_set_font(mbtk_window_t *win, mbtk_font_id_t fontid)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	XFontStruct *fnt;

	if ((fontid.l < 0) || (fontid.l >= xd->fonts.used))
		return -1;

	MAYBE_SKIP_RENDER(return 0);

	fnt = xd->fonts.array[fontid.l];
	if (fnt == NULL)
		return -1;

	xw->fnt = fnt;
	XSetFont(xd->display, xw->gc, fnt->fid);

	return 0;
}

static void mbtk_xlib_draw_query_text(mbtk_window_t *win, int *wo, int *asco, int *desco, char *s, int slen)
{
	xlib_win_t *xw = (xlib_win_t *)win;

	if (xw->fnt != NULL) {
		XCharStruct xcs;
		int dir, asc, desc;

		XTextExtents(xw->fnt, s, slen, &dir, &asc, &desc, &xcs);
/*		*wo = XTextWidth(xw->fnt, s, slen);*/
		*wo = xcs.width;
		*asco = asc;
		*desco = desc;
	}
	else {
		*wo = slen << 3;
		*asco = *desco = 5;
	}
}

static long mbtk_xlib_text_width(xlib_win_t *xw, char *s, long len, long maxlen)
{
	if (len == 0)
		return 0;

	if (len > maxlen)
		len = maxlen;

	return XTextWidth(xw->fnt, s, len);
}

/* Figure the character position for a given pixel pos; needs bisection
   because there's no better API available */
static long mbtk_xlib_text_bisect_for_px(xlib_win_t *xw, long pxpos, char *s, int slen, long *x1, long *x2)
{
	long i, lowx, lowi, highx, highi;

	if (pxpos <= 0) {
		*x1 = 0;
		*x2 = mbtk_xlib_text_width(xw, s, 1, slen);
		return 0;
	}

	lowx = lowi = 0;
	highi = slen;
	highx = LONG_MAX;

	for(i = highi/2;;) {
		long x = mbtk_xlib_text_width(xw, s, i, slen);
		if (x == pxpos) {
			*x1 = x;
			*x2 = mbtk_xlib_text_width(xw, s, i+1, slen);
			return i;
		}
		if (x < pxpos) {
			if (x > lowx) {
				lowi = i;
				lowx = x;
			}
		}
		if (x > pxpos) {
			if (x < highx) {
				highi = i;
				highx = x;
			}
		}
		if (lowi + 1 == highi) {
			*x1 = lowx;
			*x2 = highx;
			return highi-1;
		}
		i = (lowi+highi) / 2;
		if (i == lowi)
			i++;
	}
}

static long mbtk_xlib_draw_query_text_pos(mbtk_window_t *win, long *char_start_px, long *char_end_px, long pxpos, long delta_chr, char *s, int slen)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	long chpos, x1, x2;

	chpos = mbtk_xlib_text_bisect_for_px(xw, pxpos, s, slen, &x1, &x2);
	if (delta_chr == 0) {
		*char_start_px = x1+1;
		*char_end_px = x2+1;
		return chpos;
	}

	if (delta_chr > 0) {
		chpos += delta_chr;
		if (chpos > slen)
			chpos = slen;
	}
	else {
		chpos--;
		if (chpos < 0)
			return -1;
	}

	if (chpos > 0)
		*char_start_px = mbtk_xlib_text_width(xw, s, chpos, slen)+1;
	else
		*char_start_px = 0;

	if (chpos < slen)
		*char_end_px = mbtk_xlib_text_width(xw, s, chpos+1, slen)+1;
	else
		*char_end_px = *char_start_px+1;

	return chpos;

}

static void mbtk_xlib_draw_text(mbtk_window_t *win, int x, int y, char *s, int slen, void **cache)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;

	MAYBE_SKIP_RENDER(return);

	if (xw->fnt != NULL)
		XDrawString(xd->display, xw->window, xw->gc, x + OX, y + OY, s, slen);
}

static int mbtk_xlib_draw_import_pixmap(mbtk_display_t *display, mbtk_pixmap_t *dst, unsigned char *pixels, int width, int height, int has_alpha)
{
	xlib_display_t *xd = (xlib_display_t *)display;
	Visual *vis = DefaultVisual(xd->display, 0);

	dst->bedata.p = dst->bedata.raw = NULL; /* make sure to keep these invalid if conversion fails */
	return create_ximage(dst, xd->display, DefaultRootWindow(xd->display), vis, xd->depth, pixels, width, height, has_alpha);
}

static void mbtk_xlib_draw_free_pixmap(mbtk_display_t *display, mbtk_pixmap_t *px)
{
	xlib_display_t *xd = (xlib_display_t *)display;
	XImage *i = px->bedata.p;
	Pixmap m = px->bedata.ul;

	if (m != 0)
		XFreePixmap(xd->display, m);

	XDestroyImage(i);

/*	free(px->bedata.raw); - XDestroyImage() does this */
	px->bedata.p = px->bedata.raw = NULL;
}

static mbtk_px_id_t mbtk_xlib_draw_bind_pixmap(mbtk_window_t *win, mbtk_pixmap_t *px)
{
	mbtk_px_id_t pxid;
	pxid.p = px;
	pxid.l = 1; /* make it look valid */
	return pxid;
}

static void mbtk_xlib_draw_pixmap(mbtk_window_t *win, int x, int y, mbtk_px_id_t pxid)
{
	xlib_win_t *xw = (xlib_win_t *)win;
	xlib_display_t *xd = (xlib_display_t *)win->display;
	mbtk_pixmap_t *px = pxid.p;
	XImage *img = px->bedata.p;
	Pixmap mask = px->bedata.ul;

	MAYBE_SKIP_RENDER(return);

	if (mask != 0) {
		XSetClipMask(xd->display, xw->gc, mask);
		XSetClipOrigin(xd->display, xw->gc, x + OX, y + OY);
		XPutImage(xd->display, xw->window, xw->gc, img, 0, 0, x + OX, y + OY, px->w, px->h);
		XSetClipMask(xd->display, xw->gc, None);
	}
	else
		XPutImage(xd->display, xw->window, xw->gc, img, 0, 0, x + OX, y + OY, px->w, px->h);
}


static void mbtk_xlib_draw_flush(mbtk_window_t *win)
{
	xlib_display_t *xd = (xlib_display_t *)win->display;
	xlib_win_t *xw = (xlib_win_t *)win;

	if (win->in_widget != NULL) {
		if (win->native_draw_post != NULL)
			win->native_draw_post(win, &xw->window);
		return;
	}

	XFlush(xd->display);
}


static void mbtk_xlib_draw_start(mbtk_window_t *win)
{
	if (win->in_widget != NULL) {
		xlib_win_t *sxw = (xlib_win_t *)win;
		sxw->skip_render = 0;
		if (win->native_draw_pre != NULL)
			sxw->skip_render = win->native_draw_pre(win, &sxw->window);
	}
}



