/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "backend.h"
#include "display.h"
#include "draw.h"
#include "util.h"
#include "arg_t.h"

int mbtk_draw_color(mbtk_window_t *win, int r, int g, int b, int a)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_color == NULL))
		return -1;

	return win->display->be->draw_color(win, r, g, b, a);
}

int mbtk_draw_color_ul(mbtk_window_t *win, unsigned long clr)
{
	return win->display->be->draw_color(win,
		(clr & 0xff000000ul) >> 24, (clr & 0x00ff0000ul) >> 16, (clr & 0x0000ff00ul) >> 8,
		(clr & 0x000000fful));
}


void mbtk_draw_fillrect(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_fillrect == NULL))
		return;

	win->display->be->draw_fillrect(win, x1, y1, x2, y2);
}


void mbtk_draw_line(mbtk_window_t *win, int x1, int y1, int x2, int y2)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_line == NULL))
		return;

	win->display->be->draw_line(win, x1, y1, x2, y2);
}

void mbtk_draw_hline(mbtk_window_t *win, int x1, int y1, int x2)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_hline == NULL))
		return;

	win->display->be->draw_hline(win, x1, y1, x2);
}

void mbtk_draw_vline(mbtk_window_t *win, int x1, int y1, int y2)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_vline == NULL))
		return;

	win->display->be->draw_vline(win, x1, y1, y2);
}


void mbtk_draw_start(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_start == NULL))
		return;

	win->display->be->draw_start(win);
}


void mbtk_draw_flush(mbtk_window_t *win)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_flush == NULL))
		return;

	win->display->be->draw_flush(win);
}


int mbtk_draw_set_font(mbtk_window_t *win, mbtk_font_id_t fontid)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_set_font == NULL))
		return -1;

	return win->display->be->draw_set_font(win, fontid);
}

void mbtk_draw_query_text(mbtk_window_t *win, int *w, int *ascent, int *descent, char *s, int slen)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_query_text == NULL))
		return;

	win->display->be->draw_query_text(win, w, ascent, descent, s, slen);
}

long mbtk_draw_query_text_pos(mbtk_window_t *win, long *char_start_px, long *char_end_px, long pxpos, long delta_chr, char *s, int slen)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_query_text_pos == NULL))
		return -1;

	if (delta_chr < -1)
		return -1;

	if (pxpos < 0)
		return -1;

	return win->display->be->draw_query_text_pos(win, char_start_px, char_end_px, pxpos, delta_chr, s, slen);
}

void mbtk_draw_text(mbtk_window_t *win, int x, int y, char *s, int slen, void **cache)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_text == NULL))
		return;

	if (cache == NULL) {
		mbtk_error("mbtk_draw_text(): cache shall not be NULL\n");
		return;
	}

	win->display->be->draw_text(win, x, y, s, slen, cache);
}

void mbtk_draw_free_text(mbtk_window_t *win, void **cache)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_free_text == NULL))
		return;

	if (cache == NULL) {
		mbtk_error("mbtk_draw_free_text(): cache shall not be NULL\n");
		return;
	}

	win->display->be->draw_free_text(win, cache);
}


mbtk_font_id_t mbtk_resolve_font(mbtk_display_t *display, int argc, char *argv[])
{
	if ((display == NULL) || (display->be == NULL) || (display->be->resolve_font == NULL)) {
		mbtk_font_id_t fid;
		fid.l = -1;
		return fid;
	}

	return display->be->resolve_font(display, argc, argv);
}


int mbtk_draw_import_pixmap(mbtk_display_t *display, mbtk_pixmap_t *dst, unsigned char *pixels, int width, int height, int has_alpha)
{
	if ((display == NULL) || (display->be == NULL) || (display->be->draw_import_pixmap == NULL))
		return -1;

	dst->w = width;
	dst->h = height;
	dst->has_alpha = has_alpha;

	return display->be->draw_import_pixmap(display, dst, pixels, width, height, has_alpha);
}

void mbtk_draw_free_pixmap(mbtk_display_t *display, mbtk_pixmap_t *px)
{
	if ((display == NULL) || (display->be == NULL) || (display->be->draw_free_pixmap == NULL))
		return;

	display->be->draw_free_pixmap(display, px);
}

mbtk_px_id_t mbtk_draw_bind_pixmap(mbtk_window_t *win, mbtk_pixmap_t *px)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_bind_pixmap == NULL)) {
		mbtk_px_id_t pxid;
		pxid.l = -1;
		return pxid;
	}
	return win->display->be->draw_bind_pixmap(win, px);
}

void mbtk_draw_unbind_pixmap(mbtk_window_t *win, mbtk_px_id_t pxid)
{
	if ((win != NULL) && (win->display != NULL) && (win->display->be != NULL) && (win->display->be->draw_unbind_pixmap != NULL))
		win->display->be->draw_unbind_pixmap(win, pxid);
	pxid.l = -1;
}

void mbtk_draw_pixmap(mbtk_window_t *win, int x, int y, mbtk_px_id_t pxid)
{
	if ((win == NULL) || (win->display == NULL) || (win->display->be == NULL) || (win->display->be->draw_pixmap == NULL))
		return;

	win->display->be->draw_pixmap(win, x, y, pxid);
}
