/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdio.h>
#include <libmbtk/global_typedefs.h>
#include <libmbtk/handler.h>
#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/hand_range.h>

static const char my_name[] = "hand_range";

typedef struct mbtk_range_handler_s {
	mbtk_handler_t hdr;
} mbtk_range_handler_t;

typedef struct mbtk_handler_range_cfg_s { /* optional: dynamic allocated per range handler to deviate from default settings */
	unsigned disable_step:1;    /* do not do stepping, only drag & drop move */
	unsigned disable_keys:1;    /* do not react on the keyboard */
} mbtk_handler_range_cfg_t;

static mbtk_handler_range_cfg_t defcfg = {0};
static mbtk_range_handler_t range_handler;
static int range_handler_inited = 0;

static void range_moved_cb(mbtk_widget_t *widget)
{
	mbtk_kw_t kw_rmoved;

	mbtk_kw_cache(kw_rmoved, "range.moved");
	mbtk_callback_noarg(widget, kw_rmoved);
}

int mbtk_handler_range_step(mbtk_widget_t *widget, int dir, int callback)
{
	mbtk_arg_t a[4];
	mbtk_kw_t kw_range_val;
	double len;

	mbtk_kw_cache(kw_range_val, "range.val");

	if (mbtk_widget_get_propa(widget, kw_range_val, a, 4) != 4)
		return -1;

	len = a[3].v.d - a[2].v.d;

	if (dir < 0)
		len = -len;

	a[2].v.d += len;
	a[3].v.d += len;

	if (mbtk_widget_set_propa(widget, kw_range_val, a, 4) != 4)
		return -1;

	if (callback)
		range_moved_cb(widget);

	return 0;
}

int mbtk_handler_range_jump_to_end(mbtk_widget_t *widget, int dir, int callback)
{
	mbtk_arg_t a[4];
	mbtk_kw_t kw_range_val;

	mbtk_kw_cache(kw_range_val, "range.val");

	if (mbtk_widget_get_propa(widget, kw_range_val, a, 4) != 4)
		return -1;

	if (dir < 0)
		a[2].v.d = a[0].v.d;
	else
		a[2].v.d = a[1].v.d;

	a[3].type = MBTK_INVALID;

	if (mbtk_widget_set_propa(widget, kw_range_val, a, 4) != 4)
		return -1;

	if (callback)
		range_moved_cb(widget);

	return 0;
}

static mbtk_event_handled_t range_event(mbtk_widget_t *widget, mbtk_event_t *ev)
{
	mbtk_arg_t a[4], h;
	mbtk_kw_t kw_range_coords, kw_range_horiz;
	mbtk_handler_range_cfg_t *cfg = widget->handler_data;

	mbtk_kw_cache(kw_range_coords, "range.coords");
	mbtk_kw_cache(kw_range_horiz, "range.horiz");

	switch(ev->type) {
		case MBTK_EV_KEY_RELEASE:
		if (!cfg->disable_keys && !cfg->disable_step) {
			int horiz = 0;
			switch(ev->data.key.sym) {
				case MBTK_KS_LEFT: case MBTK_KS_RIGHT:
				case MBTK_KS_UP: case MBTK_KS_DOWN:
				case MBTK_KS_HOME: case MBTK_KS_END:
					if (mbtk_widget_get_prop(widget, kw_range_horiz, &h) != 1)
						break;
					horiz = h.v.l;
					break;
				default: break;
			}

			switch(ev->data.key.sym) {
				case MBTK_KS_LEFT:   if (horiz)  mbtk_handler_range_step(widget, -1, 1); return MBTK_EVENT_HANDLED;
				case MBTK_KS_RIGHT:  if (horiz)  mbtk_handler_range_step(widget, +1, 1); return MBTK_EVENT_HANDLED;
				case MBTK_KS_UP:     if (!horiz) mbtk_handler_range_step(widget, -1, 1); return MBTK_EVENT_HANDLED;
				case MBTK_KS_DOWN:   if (!horiz) mbtk_handler_range_step(widget, +1, 1); return MBTK_EVENT_HANDLED;
				case MBTK_KS_HOME:   mbtk_handler_range_jump_to_end(widget, -1, 1); return MBTK_EVENT_HANDLED;
				case MBTK_KS_END:    mbtk_handler_range_jump_to_end(widget, +1, 1); return MBTK_EVENT_HANDLED;
				default: break;
			}
			break;
		}

		case MBTK_EV_MOUSE_PRESS:
		{
			int mc, r1, r2, horiz;

			if (ev->data.mouse.button != MBTK_MB_LEFT)
				break;

			if (mbtk_widget_get_propa(widget, kw_range_coords, a, 4) != 4)
				break;
			if (mbtk_widget_get_prop(widget, kw_range_horiz, &h) != 1)
				break;
			horiz = h.v.l;

			if (horiz) {
				mc = ev->data.mouse.x;
				r1 = a[2].v.c.x;
				r2 = a[3].v.c.x;
			}
			else {
				mc = ev->data.mouse.y;
				r1 = a[2].v.c.y;
				r2 = a[3].v.c.y;
			}

			if ((mc >= r1) && (mc <= r2)) {
				/* clicked in curr */
				mbtk_window_set_focus(widget->win, widget, 1);
				mbtk_widget_set_activated(widget, 1);
				widget->win->origin = widget;
				widget->handler_static_data.l = mc - r1;
				return MBTK_EVENT_HANDLED;
			}

			if (cfg->disable_step)
				break;

			if (mc < r1) {
				mbtk_handler_range_step(widget, -1, 1);
				return MBTK_EVENT_HANDLED;
			}
			if (mc > r2) {
				mbtk_handler_range_step(widget, +1, 1);
				return MBTK_EVENT_HANDLED;
			}
			break;
		}
		case MBTK_EV_MOUSE_RELEASE:
			if (ev->data.mouse.button == MBTK_MB_LEFT) {
				if (widget->flags.activated)
					range_moved_cb(widget);
				mbtk_widget_set_activated(widget, 0);
				return MBTK_EVENT_HANDLED;
			}
			break;

		case MBTK_EV_MOUSE_MOTION:
			if (widget->win->origin != NULL) {  /* grabbed within the range for a drag&drop move */
				int horiz;

				if (mbtk_widget_get_propa(widget, kw_range_coords, a, 4) != 4)
					break;
				if (mbtk_widget_get_prop(widget, kw_range_horiz, &h) != 1)
					break;
				horiz = h.v.l;

				if (horiz)
					a[2].v.c.x = ev->data.mouse.x - widget->handler_static_data.l;
				else
					a[2].v.c.y = ev->data.mouse.y - widget->handler_static_data.l;

				a[3].type = MBTK_INVALID; /* do not change range length */

				mbtk_widget_set_propa(widget, kw_range_coords, a, 4);
				range_moved_cb(widget);
				return MBTK_EVENT_HANDLED;
			}
			break;

		case MBTK_EV_FOCUS_ENTER:
			mbtk_widget_set_activated(widget, (widget == widget->win->origin) && mbtk_display_get_mouse_button(widget->win->display, MBTK_MB_LEFT));
			return MBTK_EVENT_HANDLED;
		case MBTK_EV_FOCUS_LEAVE:
			mbtk_widget_set_activated(widget, 0);
			return MBTK_EVENT_HANDLED;

		default:;
	}

	return MBTK_EVENT_NOT_HANDLED;
}

static void range_widget_remove(mbtk_widget_t *widget)
{
	assert(widget->handler->name == my_name);

	if (widget->handler_data != &defcfg)
		free(widget->handler_data);

	widget->handler_data = NULL;
}


void mbtk_handler_range_install(mbtk_widget_t *widget)
{
	if (widget->handler != NULL)
		mbtk_widget_del_handler(widget);

	if (!range_handler_inited) {
		range_handler.hdr.name = my_name;
		range_handler.hdr.event = range_event;
		range_handler.hdr.widget_remove = range_widget_remove;
		range_handler_inited = 1;
	}
	widget->handler = &range_handler.hdr;
	widget->handler_data = &defcfg;
}

void mbtk_handler_range_cfg_step(mbtk_widget_t *widget, int enable)
{
	mbtk_handler_range_cfg_t *cfg = widget->handler_data;
	int disable = !enable;

	assert(widget->handler->name == my_name);

	if (cfg->disable_step == disable)
		return;

	if (cfg == &defcfg)
		cfg = widget->handler_data = calloc(sizeof(mbtk_handler_range_cfg_t), 1);

	cfg->disable_step = disable;
}

void mbtk_handler_range_cfg_keys(mbtk_widget_t *widget, int enable)
{
	mbtk_handler_range_cfg_t *cfg = widget->handler_data;
	int disable = !enable;

	assert(widget->handler->name == my_name);

	if (cfg->disable_keys == disable)
		return;

	if (cfg == &defcfg)
		cfg = widget->handler_data = calloc(sizeof(mbtk_handler_range_cfg_t), 1);

	cfg->disable_keys = disable;
}
