/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#ifndef LIBMBTK_WIDGET_MENU_H
#define LIBMBTK_WIDGET_MENU_H

#include <libmbtk/config.h>
#include <libmbtk/widget.h>
#include <libmbtk/widget_inlines.h>
#include <libmbtk/widget/box.h>
#include <libmbtk/widget/hand_hovradio.h>

/* A VBOX of menu items, each item being a widget, typically a label or a
   HBOX of app supplied widgets. Layout widget. */

typedef struct mbtk_menu_s {
	mbtk_widget_t w;            /* ->win is NULL or the primary menu window, the only one open menu-window of the menu, starting from it's parent_item */
	mbtk_widget_t *parent_item; /* a widget in mbtk_menu_t or mbtk_menu_bar_t */
	vtp0_t submenus;
	mbtk_handler_hovradio_t radio;

	struct {                    /* there can be only one torn-off instance */
	 mbtk_window_t *win;        /* host window when open, NULL otherwise */
	} tear_off;
} mbtk_menu_t;

/* Create a new, empty menu  */
mbtk_menu_t *mbtk_menu_new(mbtk_menu_t *caller_alloced);

/* Open or close the primary menu window. Parwin is the "parent window": the
   window the popup is opened for. */
void mbtk_menu_open_dropdown(mbtk_menu_t *menu, long screen_x, long screen_y);
void mbtk_menu_open_popup(mbtk_menu_t *menu, mbtk_window_t *parwin, long screen_x, long screen_y);
void mbtk_menu_close(mbtk_menu_t *menu);

/* This closes menu and all its parents; if menu is open from a menu bar, it
   also makes the menu bar lose cursor. Typically ran after a succesful menu
   item invocation or cancel */
void mbtk_menu_close_all(mbtk_menu_t *menu);

/* Append an item (a row) at the end of the list:
   - the generic version accepts any widget for the row (should be a hbox).
   - the _stdrow variant creates a standard row hbox 
   - the _label version creates an stdrow with a label from string;
     returns row box
   - the _label_submenu is the same as _label but also adds submenu indication
     on the right (suffix)
   - the _label_full variant optionally creates a checkbox for prefix and a
     hotkey label for suffix (may be NULL); returns row box
*/
void mbtk_menu_append(mbtk_menu_t *menu, mbtk_widget_t *row);
mbtk_box_t *mbtk_menu_append_stdrow(mbtk_menu_t *menu, mbtk_widget_t *prefix, mbtk_widget_t *entry, mbtk_widget_t *suffix);
mbtk_box_t *mbtk_menu_append_label(mbtk_menu_t *menu, const char *item_name);
mbtk_box_t *mbtk_menu_append_label_submenu(mbtk_menu_t *menu, const char *item_name);
mbtk_box_t *mbtk_menu_append_label_full(mbtk_menu_t *menu, int chkbox, const char *item_name, const char *hotkey);

/* Append a horizontal separator */
mbtk_widget_t *mbtk_menu_append_sep(mbtk_menu_t *menu);

/* Low level stdrow creation; returns an stdrow hbox that can be added in a
   menu. An stdrow contains: a
    - a prefix (e.g. for a checkbox) or NULL,
    - the actual entry, which should normally be a label or picture
    - a suffix (e.g. for hotkey or submenu mark) or NULL
   If prefix or suffix is NULL, an empty vbox is added instead.
   Returns newly allocated floating row box.
*/
mbtk_box_t *mbtk_menu_build_stdrow(mbtk_widget_t *prefix, mbtk_widget_t *entry, mbtk_widget_t *suffix);
mbtk_box_t *mbtk_menu_build_label_full_stdrow(int chkbox, const char *item_name, const char *hotkey);

/* Build and return a stdrow for a submenu with a label and a right arrow */
mbtk_box_t *mbtk_menu_build_label_submenu_stdrow(const char *item_name);


/* Insert an item (a row) at an existing row in a menu */
void mbtk_menu_add_at(mbtk_widget_t *existing_row, mbtk_widget_t *new_row, int ins_before);


/* Make menu_item open submenu; the same submenu can be attahced to multiple
   items even in different menus. Menu_item must be a first level widget
   within a menu, typically a hbox for an stdrow (as returned by the call
   created the menu item, e.g. mbtk_menu_append_*()). Returns 0 on success */
int mbtk_menu_item_attach_submenu(mbtk_widget_t *menu_item, mbtk_menu_t *submenu);

/* Helpers to bind a callback to a menu item. The stdrow variant does the same
   but saves a cast to box on the caller's side */
void mbtk_menu_item_callback_invoked(mbtk_widget_t *row, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data);
void mbtk_menu_stdrow_callback_invoked(mbtk_box_t *row, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data);

/* Manipulate the checkbox on an stdrow (or any row that is a hbox with the
   first child being a checkbox). No callback made on checkbox state change.
   The ones returning int are returning the (new) state of the checkbox or
   -1 on error */
int mbtk_menu_stdrow_checkbox_get(mbtk_box_t *row);
void mbtk_menu_stdrow_checkbox_set(mbtk_box_t *row, int checked);
int mbtk_menu_stdrow_checkbox_toggle(mbtk_box_t *row);


MBTK_WIDGET_INLINES(menu)

#endif
