/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <math.h>
#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/draw.h>
#include <libmbtk/widget/scrollbar.h>
#include <libmbtk/widget/hand_range.h>
#include <libmbtk/widget/Common.h>


static void mbtk_scrollbar_imp_destroy(mbtk_widget_t *w) { }

#define FILLSIZE 8
#define DEFAULT_THICKNESS 16

static int mbtk_scrollbar_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	long l, r, t, b;
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
	mbtk_widget_flags_t flags = w->flags;
	mbtk_scrollbar_t *scb = (mbtk_scrollbar_t *)w;

	flags.activated = 1;
	style->min_box_frame_size(win, sp, flags, &l, &r, &t, &b);

	w->hvbox.fr_l = l; w->hvbox.fr_r = r;
	w->hvbox.fr_t = t; w->hvbox.fr_b = b;
	if (scb->horiz) {
		*mx = l+r+FILLSIZE; *my = t+b+scb->thickness;
	}
	else {
		*mx = l+r+scb->thickness; *my = t+b+FILLSIZE;
	}

	return 0;
}

static void mbtk_scrollbar_calc(mbtk_widget_t *w, int draw, mbtk_arg_t *crds)
{
	mbtk_scrollbar_t *scb = (mbtk_scrollbar_t *)w;
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
	mbtk_widget_flags_t flags = w->flags;
	long l, r, t, b, bx1, by1, bx2, by2, bwi, bhe;
	long x1 = w->hvbox.x, y1 = w->hvbox.y, wi = w->hvbox.w, he = w->hvbox.h-1;
	double pos, len;

	/* clamp and normalize between 0 and 1 */
	if (scb->max > scb->min) {
		pos = scb->curr;
		len = scb->len;

		if (len > scb->max - scb->min)
			len = scb->max - scb->min;
		else if (len < 0)
			len = 0;

		if (pos < scb->min) pos = scb->min;
		if (pos+len > scb->max) pos = scb->max-len;
		pos = (pos - scb->min) / (scb->max - scb->min);
		len /= (scb->max - scb->min);
	}
	else
		pos = len = 0;


	flags.activated = 1;
	style->min_box_frame_size(win, sp, flags, &l, &r, &t, &b);
	if (draw)
		style->draw_bg_rect(win, sp, flags, x1, y1, wi, he);

	/* tune dimensions with the frame for the bar */
	bwi = wi-(l+r);
	bhe = he-(t+b);
	if (scb->horiz) {
		bx1 = x1 + l + ((double)bwi * pos);    by1 = y1+t;
		bx2 = bx1 + floor((double)bwi * len);  by2 = by1+bhe-1;
	}
	else {
		bx1 = x1 + l;     by1 = y1+t + ((double)bhe * pos);
		bx2 = bx1+bwi-1;  by2 = by1 + floor((double)bhe * len);
	}

	if (draw) {
		flags.activated = w->flags.activated;
		style->draw_bg_rect(win, sp, flags, bx1, by1, bx2-bx1+1, by2-by1+1);

		flags.activated = 0;
		style->draw_box_frame(win, sp, flags, bx1, by1, bx2-bx1+1, by2-by1+1);


		flags.activated = 1;
		style->draw_box_frame(win, sp, flags, x1, y1, wi, he);
	}

	if (crds != 0) {
		crds[0].type = MBTK_COORD; crds[0].v.c.x = x1;    crds[0].v.c.y = y1;
		crds[1].type = MBTK_COORD; crds[1].v.c.x = x1+wi; crds[1].v.c.y = y1+he;
		crds[2].type = MBTK_COORD; crds[2].v.c.x = bx1;   crds[2].v.c.y = by1;
		crds[3].type = MBTK_COORD; crds[3].v.c.x = bx2;   crds[3].v.c.y = by2;
	}
}

static void mbtk_scrollbar_fallback_draw(mbtk_widget_t *w)
{
	mbtk_scrollbar_calc(w, 1, NULL);
}

static void mbtk_scrollbar_fallback_draw_overlay(mbtk_widget_t *w)
{
	mbtk_arg_t a[4];

	mbtk_scrollbar_calc(w, 0, a);

	if (((a[3].v.c.x - a[2].v.c.x) > 8) && ((a[3].v.c.y - a[2].v.c.y) > 8)) {
		mbtk_window_t *win = w->win;
		const mbtk_style_t *style = mbtk_widget_style(w);
		mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
		mbtk_widget_flags_t flags = w->flags;

		flags.activated = 0;
		style->draw_box_overlay(win, sp, flags, a[2].v.c.x, a[2].v.c.y, a[3].v.c.x - a[2].v.c.x + 1, a[3].v.c.y - a[2].v.c.y + 1);
	}
	else
		mbtk_comm_frame_fallback_draw_overlay(w);
}

static int mbtk_scrollbar_get_propa(mbtk_widget_t *w, mbtk_kw_t key, mbtk_arg_t *dst, int arrlen)
{
	mbtk_kw_t kw_range_val, kw_range_coords, kw_range_horiz;
	mbtk_scrollbar_t *scb = (mbtk_scrollbar_t *)w;

	mbtk_kw_cache(kw_range_val, "range.val");
	mbtk_kw_cache(kw_range_coords, "range.coords");
	mbtk_kw_cache(kw_range_horiz, "range.horiz");

	if ((key == kw_range_val) && (arrlen == 4)) {
		dst[0].type = MBTK_DOUBLE; dst[0].v.d = scb->min;
		dst[1].type = MBTK_DOUBLE; dst[1].v.d = scb->max;
		dst[2].type = MBTK_DOUBLE; dst[2].v.d = scb->curr;
		dst[3].type = MBTK_DOUBLE; dst[3].v.d = scb->curr + scb->len;
		return 4;
	}

	if ((key == kw_range_coords) && (arrlen == 4)) {
		mbtk_scrollbar_calc(w, 0, dst);
		return 4;
	}

	if ((key == kw_range_horiz) && (arrlen == 1)) {
		dst[0].type = MBTK_BOOL; dst[0].v.b = scb->horiz;
		return 1;
	}

	return -1;
}

static int mbtk_scrollbar_set_propa(mbtk_widget_t *w, mbtk_kw_t key, const mbtk_arg_t *src, int arrlen)
{
	mbtk_kw_t kw_range_val, kw_range_coords;
	mbtk_scrollbar_t *scb = (mbtk_scrollbar_t *)w;

	mbtk_kw_cache(kw_range_val, "range.val");
	mbtk_kw_cache(kw_range_coords, "range.coords");

	if ((key == kw_range_val) && (arrlen == 4)) {
		if ((src[0].type != MBTK_DOUBLE) || (src[1].type != MBTK_DOUBLE) || (src[2].type != MBTK_DOUBLE))
			return -1;

		scb->min = src[0].v.d;
		scb->max = src[1].v.d;
		scb->curr = src[2].v.d;

		if (src[3].type == MBTK_DOUBLE) {
			scb->len = src[3].v.d - src[2].v.d;

			clamp_and_set_both:;
			if (scb->len > scb->max - scb->min)
				scb->len = scb->max - scb->min;
			else if (scb->len < 0)
				scb->len = 0;
		}

		clamp_and_set_curr:;
		if (scb->curr < scb->min)
			scb->curr = scb->min;
		if (scb->curr + scb->len > scb->max)
			scb->curr = scb->max - scb->len;
		mbtk_widget_needs_redraw(w);
		return 4;
	}

	if ((key == kw_range_coords) && (arrlen == 4)) {
		double pxlen;

		if (scb->horiz) {
			double x1 = src[0].v.c.x;
			pxlen = src[1].v.c.x - src[0].v.c.x;
			scb->curr = (src[2].v.c.x - x1) / pxlen * (scb->max - scb->min) + scb->min;
		}
		else {
			double y1 = src[0].v.c.y;
			pxlen = src[1].v.c.y - src[0].v.c.y;
			scb->curr = (src[2].v.c.y - y1) / pxlen * (scb->max - scb->min) + scb->min;
		}

		if (src[3].type != MBTK_COORD)
			goto clamp_and_set_curr; /* do not change length */

		if (scb->horiz)
			scb->len = (double)((src[3].v.c.x - src[2].v.c.x) + 0.5) / pxlen * (scb->max - scb->min);
		else
			scb->len = (double)((src[3].v.c.y - src[2].v.c.y) + 0.5) / pxlen * (scb->max - scb->min);
		goto clamp_and_set_both;
	}

	return -1;
}


static const mbtk_widget_imp_t mbtk_scrollbar_imp = {
	mbtk_scrollbar_imp_destroy,

	mbtk_scrollbar_fallback_minsize,
	mbtk_scrollbar_fallback_draw,
	mbtk_scrollbar_fallback_draw_overlay,

	mbtk_scrollbar_get_propa,
	mbtk_scrollbar_set_propa,

	NULL, /* post_assign */
	NULL, /* pre_unassign */

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	NULL  /* pre_child_unlink */
};

mbtk_scrollbar_t *mbtk_scrollbar_new(mbtk_scrollbar_t *caller_alloced, int horiz)
{
	mbtk_scrollbar_t *scrollbar;
	mbtk_kw_t kw_scrollbar;

	mbtk_kw_cache(kw_scrollbar, "scrollbar");

	mbtk_create_wt(scrollbar, kw_scrollbar, &mbtk_scrollbar_imp, caller_alloced);
	scrollbar->min = 0.0;
	scrollbar->max = 1.0;
	scrollbar->curr = 0.0;
	scrollbar->len = 0.2;
	scrollbar->w.hvbox.type = horiz ? HVBOX_HBOX : HVBOX_VBOX; /* don't count padding be part of the widget */
	scrollbar->horiz = horiz;
	scrollbar->thickness = DEFAULT_THICKNESS;

	mbtk_handler_range_install(&scrollbar->w);

	return scrollbar;
}

int mbtk_scrollbar_set_curr(mbtk_scrollbar_t *prg, double curr, double len, int clamp)
{
	if (len > prg->max - prg->min) {
		if (!clamp)
			return -1;
		len = prg->max - prg->min;
	}

	if (curr < prg->min) {
		if (!clamp)
			return -1;
		prg->curr = prg->min;
	}
	else if (curr+len > prg->max) {
		if (!clamp)
			return -1;
		prg->curr = prg->max-len;
	}
	else
		prg->curr = curr;

	prg->len = len;

	mbtk_widget_needs_redraw(&prg->w);
	return 0;
}

void mbtk_scrollbar_callback_range_moved(mbtk_scrollbar_t *b, mbtk_event_handled_t (*cb)(mbtk_widget_t *w, mbtk_kw_t id, void *user_data), void *user_data)
{
	mbtk_kw_t kw_range_moved;
	mbtk_kw_cache(kw_range_moved, "range.moved");
	mbtk_callback_set_noarg(&b->w, kw_range_moved, cb, user_data);
}

