/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/util.h>
#include <libmbtk/handler.h>
#include <libmbtk/callback_imp.h>
#include <libmbtk/widget/hand_ev2cb.h>
#include <libmbtk/widget/scrolled.h>

static void mbtk_scrolled_imp_destroy(mbtk_widget_t *w)
{
	mbtk_scrolled_t *scrolled = (mbtk_scrolled_t *)w;

	if (scrolled->root != NULL) {
		mbtk_widget_destroy(scrolled->root);
		scrolled->root = NULL;
	}
}


static int mbtk_scrolled_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	*mx = *my = 0; /* rely on user set minimums */
	return 0;
}

static void mbtk_scrolled_assign_root_widget(mbtk_scrolled_t *scrolled, mbtk_widget_t *widget)
{
	static mbtk_handler_t dummy;

	if (widget->handler == NULL)
		widget->handler = &dummy; /* so that set_root_widget() won't set the handler */

	mbtk_window_set_root_widget(scrolled->subwin, widget);

	if (widget->handler == &dummy)
		widget->handler = NULL;
}

static void mbtk_scrolled_post_assign(mbtk_widget_t *w)
{
	mbtk_scrolled_t *scrolled = (mbtk_scrolled_t *)w;

	assert(scrolled->subwin == NULL);
	scrolled->subwin = mbtk_subwindow_new(w->win->display, &scrolled->w, 0, 0, 0, 0);
	scrolled->subwin->scrolled = 1;
	mbtk_subwindow_assign(scrolled->subwin, w->render_win);
	if (scrolled->root != NULL)
		mbtk_scrolled_assign_root_widget(scrolled, scrolled->root);
}

static void mbtk_scrolled_pre_unassign(mbtk_widget_t *w)
{
	mbtk_scrolled_t *scrolled = (mbtk_scrolled_t *)w;

	if (scrolled->root != NULL)
		mbtk_widget_unlink(scrolled->root);

	mbtk_subwindow_unassign(scrolled->subwin);
	mbtk_subwindow_free(scrolled->subwin);
	scrolled->subwin = NULL;
}


static void mbtk_scrolled_fallback_draw(mbtk_widget_t *w)
{
	mbtk_scrolled_t *scrolled = (mbtk_scrolled_t *)w;

	/* ensure subwindow geometry matches widget's */
	mbtk_subwindow_geometry(scrolled->subwin, w->hvbox.x, w->hvbox.y, MBTK_MAX(w->hvbox.w, 1), MBTK_MAX(w->hvbox.h, 1));

	scrolled->subwin->draw_dirty = 1;
	mbtk_win_redraw(scrolled->subwin);
}

static void mbtk_scrolled_fallback_draw_overlay(mbtk_widget_t *w)
{
}



static const mbtk_widget_imp_t mbtk_scrolled_imp = {
	mbtk_scrolled_imp_destroy,

	mbtk_scrolled_fallback_minsize,
	mbtk_scrolled_fallback_draw,
	mbtk_scrolled_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	mbtk_scrolled_post_assign,
	mbtk_scrolled_pre_unassign,

	NULL, /* pre_appear */
	NULL, /* post_disappear */

	NULL  /* pre_child_unlink */
};

mbtk_scrolled_t *mbtk_scrolled_new(mbtk_scrolled_t *caller_alloced)
{
	mbtk_scrolled_t *scrolled;
	mbtk_kw_t kw_scrolled;

	mbtk_kw_cache(kw_scrolled, "scrolled");

	mbtk_create_wt(scrolled, kw_scrolled, &mbtk_scrolled_imp, caller_alloced);

	scrolled->w.relative_ev_coords = 1;
	mbtk_handler_ev2cb_install(&scrolled->w);
	return scrolled;
}



void mbtk_scrolled_set_root_widget(mbtk_scrolled_t *scrolled, mbtk_widget_t *widget)
{
	mbtk_scrolled_assign_root_widget(scrolled, widget);
	scrolled->root = widget;
}


#define get_scroll_max(maxx, maxy, cntw, cnth) \
	do { \
		mbtk_widget_t *root = mbtk_window_get_root(scrolled->subwin); \
		if (root != NULL) { \
			cntw = root->hvbox.w; \
			cnth = root->hvbox.h; \
			maxx = MBTK_MAX(cntw - scrolled->subwin->w, 0); \
			maxy = MBTK_MAX(cnth - scrolled->subwin->h, 0); \
		} \
		else \
			maxx = maxy = 0; \
	} while(0)


int mbtk_scrolled_set_content_origin(mbtk_scrolled_t *scrolled, long cx0, long cy0)
{
	long maxx, maxy, cntw, cnth;

	get_scroll_max(maxx, maxy, cntw, cnth);

	if (cx0 > maxx) cx0 = maxx;
	else if (cx0 < 0) cx0 = 0;
	if (cy0 > maxy) cy0 = maxy;
	else if (cy0 < 0) cy0 = 0;

	if ((cx0 == scrolled->subwin->content_x0) && (cy0 == scrolled->subwin->content_y0))
		return 0;

	scrolled->subwin->content_x0 = cx0;
	scrolled->subwin->content_y0 = cy0;
	mbtk_widget_needs_redraw(&scrolled->w);

	return 1;
}

void mbtk_scrolled_get_range(mbtk_scrolled_t *scrolled, mbtk_arg_t dst[4], int horiz)
{
	long maxx, maxy, cntw, cnth;

	get_scroll_max(maxx, maxy, cntw, cnth);

	(void)maxx; (void)maxy;

	dst[0].type = MBTK_DOUBLE; dst[0].v.d = 0;
	if (horiz) {
		dst[1].type = MBTK_DOUBLE; dst[1].v.d = cntw;
		dst[2].type = MBTK_DOUBLE; dst[2].v.d = scrolled->subwin->content_x0;
		dst[3].type = MBTK_DOUBLE; dst[3].v.d = dst[2].v.d + MBTK_MIN(scrolled->subwin->w, cntw);
	}
	else {
		dst[1].type = MBTK_DOUBLE; dst[1].v.d = cnth;
		dst[2].type = MBTK_DOUBLE; dst[2].v.d = scrolled->subwin->content_y0;
		dst[3].type = MBTK_DOUBLE; dst[3].v.d = dst[2].v.d + MBTK_MIN(scrolled->subwin->h, cnth);
	}
}
