/*  miniboxtk - small GUI toolkit/widget system
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    (Supported by NLnet NGI0 PET Fund in 2022)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Contact:
      Project page: http://repo.hu/projects/miniboxtk
      contact lead developer: http://www.repo.hu/projects/miniboxtk/contact.html
*/

#include <stdlib.h>
/*#include <stdio.h>*/
#include <string.h>
#include <stddef.h>
#include <assert.h>
#include <libmbtk/draw.h>
#define MBTK_NO_INCLUDE_WIDGETS
#include <libmbtk/widget.h>
#include <libmbtk/util.h>
#include <libmbtk/widget_imp.h>
#include <libmbtk/widget/textline.h>


static void mbtk_textline_imp_destroy(mbtk_widget_t *w)
{
	mbtk_textline_t *textline = (mbtk_textline_t *)w;

	gds_uninit(&textline->str);
}


/* Recalculate textline->width */
static int mbtk_textline_line_recalc(mbtk_textline_t *textline)
{
	mbtk_window_t *win = textline->w.win;
	int w, asc, dsc;

	if (win == NULL)
		return -1;

	mbtk_draw_query_text(win, &w, &asc, &dsc, textline->str.array, textline->str.used);
	textline->font_asc = asc;
	textline->font_dsc = dsc;
	textline->width = w;

	textline->width_valid = 1;
	return 0;
}

static int mbtk_textline_fallback_minsize(mbtk_widget_t *w, long *mx, long *my)
{
	mbtk_textline_t *textline = (mbtk_textline_t *)w;
	mbtk_window_t *win = textline->w.win;

	if (textline->fid.l == -1)
		mbtk_draw_set_font(win, win->display->default_fid);
	else
		mbtk_draw_set_font(win, textline->fid);

	if (!textline->width_valid) {
		if (mbtk_textline_line_recalc(textline) != 0) {
			mbtk_error("mbtk_textline_draw(): invalid font\n");
			return -1;
		}
	}

	*mx = textline->width;
	*my = textline->font_asc + textline->font_dsc;

	return 0;
}

static void mbtk_textline_fallback_draw(mbtk_widget_t *w)
{
	mbtk_window_t *win = w->win;
	const mbtk_style_t *style = mbtk_widget_style(w);
	mbtk_styleprop_t *sp = mbtk_styleprop_get(w);
	mbtk_textline_t *textline = (mbtk_textline_t *)w;
	long x1 = textline->w.hvbox.x, y1 = textline->w.hvbox.y, x, y;

	style->set_fg_color(win, sp, w->flags);
	if (textline->fid.l == -1)
		mbtk_draw_set_font(win, win->display->default_fid);
	else
		mbtk_draw_set_font(win, textline->fid);

	if (!textline->width_valid) {
		if (mbtk_textline_line_recalc(textline) != 0) {
			mbtk_error("mbtk_textline_draw(): invalid font\n");
			return;
		}
	}

	x = x1;
	y = y1 + textline->font_asc;

	mbtk_draw_text(win, x, y, textline->str.array, textline->str.used, &textline->cookie);
}

static void mbtk_textline_fallback_draw_overlay(mbtk_widget_t *w)
{
	mbtk_textline_t *textline = (mbtk_textline_t *)w;
	mbtk_window_t *win = w->win;
	long y1, y2;
	const mbtk_style_t *style;
	mbtk_styleprop_t *sp;

	if (textline->cursor < 0)
		return;

	style = mbtk_widget_style(w);
	sp = mbtk_styleprop_get(w);
	y1 = w->hvbox.y + 1;
	y2 = w->hvbox.y + w->hvbox.h - 1;


	style->set_fg_color(win, sp, w->flags);
	mbtk_draw_vline(win, textline->cursor_x, y1, y2);

	if (textline->overwrite) {
		mbtk_draw_vline(win, textline->cursor_x2, y1, y2);
		mbtk_draw_hline(win, textline->cursor_x, y1, textline->cursor_x2);
		mbtk_draw_hline(win, textline->cursor_x, y2, textline->cursor_x2);
	}
}

/* Drop the cache when the widget disappears: there may be a lot of textlines
   in a scrolled window (typical for the tree-table widget) or a lot of
   them hidden (typical for the tabbed widget). Dropping cache for
   disappeared textlines helps keeping memory footprint small. */
static void mbtk_textline_post_disappear(mbtk_widget_t *w)
{
	mbtk_textline_t *textline = (mbtk_textline_t *)w;

	mbtk_draw_free_text(textline->w.win, &textline->cookie);
	textline->cookie = NULL;
}



static const mbtk_widget_imp_t mbtk_textline_imp = {
	mbtk_textline_imp_destroy,

	mbtk_textline_fallback_minsize,
	mbtk_textline_fallback_draw,
	mbtk_textline_fallback_draw_overlay,

	NULL,  /* get_prop */
	NULL,  /* set_prop */

	NULL, /* post_assign */
	NULL, /* unassign */

	NULL,
	mbtk_textline_post_disappear,

	NULL  /* pre_child_unlink */
};


mbtk_textline_t *mbtk_textline_new(mbtk_textline_t *caller_alloced, const char *str)
{
	mbtk_textline_t *textline;
	mbtk_kw_t kw_textline;

	mbtk_kw_cache(kw_textline, "textline");

	mbtk_create_wt(textline, kw_textline, &mbtk_textline_imp, caller_alloced);
	textline->fid = mbtk_invalid_font_id;
	gds_init(&textline->str);
	if (str != NULL)
		mbtk_textline_set_str(textline, str);

	textline->cursor = -1;

	return textline;
}

int mbtk_textline_cursor_set_chr(mbtk_textline_t *textline, long chpos, int delta)
{
	int overwrite = (delta == 0);
	long new_chpos, start_px, end_px;

	if ((chpos == textline->cursor) && (overwrite == textline->overwrite))
		return 0;

	if (textline->str.used != 0) {
		new_chpos = mbtk_draw_query_text_pos(textline->w.win, &start_px, &end_px, 0, chpos, textline->str.array, textline->str.used);
		if (new_chpos < 0)
			return -1;
	}
	else
		new_chpos = start_px = end_px = 0; /* corner case: keep cursor even in empty string */

	textline->cursor = new_chpos;
	textline->overwrite = overwrite;
	textline->cursor_x = start_px;
	textline->cursor_x2 = end_px;

	/* corner case: overwrite cursor after the last char should have non-zero length */
	if (new_chpos >= textline->str.used)
		textline->cursor_x2 += textline->font_dsc;

/*printf("cursor_set_chr: %ld..%ld chr=%ld of %ld overw=%d\n", start_px, end_px, new_chpos, textline->str.used, overwrite);*/

	mbtk_widget_needs_redraw(&textline->w);

	return 0;
}

int mbtk_textline_cursor_set_win(mbtk_textline_t *textline, long x)
{
	long new_chpos, start_px, end_px, dx = MBTK_MAX(x - textline->w.hvbox.x, 0);
	new_chpos = mbtk_draw_query_text_pos(textline->w.win, &start_px, &end_px, dx, 0, textline->str.array, textline->str.used);

	/* in insert mode find the closer glyph boundary: if clicked right to
	   the middle, go one position right */
	if (!textline->overwrite && (dx > (start_px+end_px)/2)) {
/*		printf("right %ld was %ld\n", new_chpos, textline->cursor);*/
		return mbtk_textline_cursor_set_chr(textline, new_chpos+1, textline->overwrite ? 0 : -1);
	}

/*	printf("left %ld != %ld\n", new_chpos, textline->cursor);*/
	if (new_chpos != textline->cursor) {
		textline->cursor = new_chpos;
		textline->cursor_x = start_px;
		textline->cursor_x2 = end_px;
		mbtk_widget_needs_redraw(&textline->w);
	}
	return 0;
}

int mbtk_textline_cursor_home(mbtk_textline_t *textline)
{
	return mbtk_textline_cursor_set_chr(textline, 0, textline->overwrite ? 0 : -1);
}

int mbtk_textline_cursor_end(mbtk_textline_t *textline)
{
	return mbtk_textline_cursor_set_chr(textline, textline->str.used, textline->overwrite ? 0 : -1);
}


void mbtk_textline_cursor_remove(mbtk_textline_t *textline)
{
	if (textline->cursor < 0)
		return;
	textline->cursor = -1;
	mbtk_widget_needs_redraw(&textline->w);
}

void mbtk_textline_set_str_len(mbtk_textline_t *textline, const char *str, long len)
{
	textline->str.used = 0;

	if (str != NULL)
		gds_append_len(&textline->str, str, len);
	else if (textline->str.array == NULL)
		gds_append_str(&textline->str, "");

	textline->cookie = NULL;

	if (mbtk_textline_cursor_set_chr(textline, textline->cursor, textline->overwrite ? 0 : 1) < 0)
		mbtk_textline_cursor_remove(textline);

	mbtk_widget_needs_redraw(&textline->w);

}

void mbtk_textline_str_changed(mbtk_textline_t *textline)
{
	textline->cookie = NULL;

	if (mbtk_textline_cursor_set_chr(textline, textline->cursor, textline->overwrite ? 0 : 1) < 0)
		mbtk_textline_cursor_remove(textline);

	textline->width_valid = 0;
	mbtk_widget_needs_redraw(&textline->w);
	mbtk_widget_changed(&textline->w, 1);
}

void mbtk_textline_set_str(mbtk_textline_t *textline, const char *str)
{
	textline->str.used = 0;

	if (str != NULL)
		gds_append_str(&textline->str, str);
	else if (textline->str.array == NULL)
		gds_append_str(&textline->str, "");

	mbtk_textline_str_changed(textline);
}

void mbtk_textline_append_str_len(mbtk_textline_t *textline, const char *str, long len)
{
	if (len == 0)
		return;
	gds_append_len(&textline->str, str, len);
	mbtk_textline_str_changed(textline);
}

void mbtk_textline_append_str(mbtk_textline_t *textline, const char *str)
{
	if ((str == NULL) || (*str == '\0'))
		return;
	gds_append_str(&textline->str, str);
	mbtk_textline_str_changed(textline);
}


void mbtk_textline_truncate(mbtk_textline_t *textline, long len)
{
	if (len >= textline->str.used) return;
	if (len < 0) len = 0;

	textline->str.used = len;
	textline->str.array[len] = '\0';
	mbtk_textline_str_changed(textline);
}


int mbtk_textline_type_str(mbtk_textline_t *textline, const char *str)
{
	int cursor = textline->cursor;

	if (textline->cursor < 0)
		return -1;

	if (textline->overwrite) {
		for(; (cursor < textline->str.used) && (*str != '\0'); cursor++,str++)
			textline->str.array[cursor] = *str;
		if (*str != 0) {
			gds_append_str(&textline->str, str);
			cursor = textline->str.used;
		}
	}
	else {
		long len = strlen(str);
		gds_insert_len(&textline->str, cursor, (char *)str, len);
		cursor += len;
	}


	mbtk_textline_cursor_set_chr(textline, cursor, textline->overwrite ? 0 : +1);

	textline->cookie = NULL;
	mbtk_textline_str_changed(textline);

	return 0;
}

static void mbtk_textline_del_(mbtk_textline_t *textline, int num, int dir)
{
	int cursor = textline->cursor;

	if (dir > 0) {
		gds_remove(&textline->str, cursor, num);
		if (cursor > textline->str.used)
			cursor = textline->str.used;
	}
	else {
		if (cursor == 0)
			return;
		cursor--;
		gds_remove(&textline->str, cursor, 1);
	}

	/* don't let the cursor disappear */
	if (cursor < 0)
		cursor = 0;

	textline->cursor = -1; /* always recalculate cursor because of the text change */
	mbtk_textline_cursor_set_chr(textline, cursor, textline->overwrite ? 0 : +1);

	textline->cookie = NULL;
	mbtk_widget_needs_redraw(&textline->w);
}

void mbtk_textline_del(mbtk_textline_t *textline, int num)
{
	mbtk_textline_del_(textline, num, +1);
}

void mbtk_textline_backspace(mbtk_textline_t *textline)
{
	mbtk_textline_del_(textline, 1, -1);
}

